from conference_scheduler import scheduler
from conference_scheduler.lp_problem import constraints


def array_violations(array, events, slots):
    """Take a schedule in array form and return any violated constraints

    Parameters
    ----------
        array : np.array
            a schedule in array form
        events : list or tuple
            of resources.Event instances
        slots : list or tuple
            of resources.Slot instances
        constraints : list or tuple
            of generator functions which each produce instances of
            resources.Constraint

    Returns
    -------
        Generator
            of a list of strings indicating the nature of the violated
            constraints
    """
    return (
        c.label
        for c in constraints.all_constraints(events, slots, array)
        if not c.condition
    )


def is_valid_array(array, events, slots):
    """Take a schedule in array form and return whether it is a valid
    solution for the given constraints

    Parameters
    ----------
        array : np.array
            a schedule in array form
        events : list or tuple
            of resources.Event instances
        slots : list or tuple
            of resources.Slot instances

    Returns
    -------
        bool
            True if array represents a valid solution
    """

    if len(array) == 0:
        return False
    violations = sum(1 for c in (array_violations(array, events, slots)))
    return violations == 0


def is_valid_solution(solution, events, slots):
    """Take a solution and return whether it is valid for the
    given constraints

    Parameters
    ----------
        solution: list or tuple
            a schedule in solution form
        events : list or tuple
            of resources.Event instances
        slots : list or tuple
            of resources.Slot instances

    Returns
    -------
        bool
            True if schedule is a valid solution
    """
    if len(solution) == 0:
        return False
    array = scheduler.solution_to_array(solution, events, slots)
    return is_valid_array(array, events, slots)


def solution_violations(solution, events, slots):
    """Take a solution and return a list of violated constraints

    Parameters
    ----------
        solution: list or tuple
            a schedule in solution form
        events : list or tuple
            of resources.Event instances
        slots : list or tuple
            of resources.Slot instances

    Returns
    -------
        Generator
            of a list of strings indicating the nature of the violated
            constraints
    """
    array = scheduler.solution_to_array(solution, events, slots)
    return array_violations(array, events, slots)


def is_valid_schedule(schedule, events, slots):
    """Take a schedule and return whether it is a valid solution for the
    given constraints

    Parameters
    ----------
        schedule : list or tuple
            a schedule in schedule form
        events : list or tuple
            of resources.Event instances
        slots : list or tuple
            of resources.Slot instances

    Returns
    -------
        bool
            True if schedule is a valid solution
    """
    if len(schedule) == 0:
        return False
    array = scheduler.schedule_to_array(schedule, events, slots)
    return is_valid_array(array, events, slots)


def schedule_violations(schedule, events, slots):
    """Take a schedule and return a list of violated constraints

    Parameters
    ----------
        schedule : list or tuple
            a schedule in schedule form
        events : list or tuple
            of resources.Event instances
        slots : list or tuple
            of resources.Slot instances

    Returns
    -------
        Generator
            of a list of strings indicating the nature of the violated
            constraints
    """
    array = scheduler.schedule_to_array(schedule, events, slots)
    return array_violations(array, events, slots)



import pulp
import numpy as np
import conference_scheduler.lp_problem as lp
from conference_scheduler.resources import ScheduledItem, Shape


# Three functions that can be called by external programs to produce the
# schedule in one of three forms:
#   solution: a generator for a list of tuples of event index and slot index
#             for each scheduled item
#   array: a numpy array with rows for events and columns for slots
#   schedule: a generator for a list of ScheduledItem instances


def solution(events, slots, objective_function=None, **kwargs):
    """Setup up the ILP problem, submit it to pulp and return the solution

    Parameters
    ----------
        events : list or tuple
            of resources.Event instances
        slots : list or tuple
            of resources.Slot instances
        objective_function: callable
            from lp_problem.objective_functions

    Returns
    -------
        Generator
            of tuples giving the event and slot index (for the given events and
            slots lists) for all scheduled items.

            e.g. for a solution where:
                event 0 is scheduled in slot 1
                event 1 is scheduled in slot 4
                event 2 is scheduled in slot 5

            the resulting generator would produce:
                [(0, 1), (1, 4), (2, 5)]
    """
    shape = Shape(len(events), len(slots))
    problem = pulp.LpProblem()
    X = lp.utils.variables(shape)

    for constraint in lp.constraints.all_constraints(events, slots, X):
        problem += constraint.condition

    if objective_function is not None:
        problem += objective_function(events=events, slots=slots, X=X,
                                      **kwargs)

    status = problem.solve()
    if status == 1:
        return (
            item for item, variable in X.items()
            if variable.value() > 0
        )
    else:
        raise ValueError('No valid solution found')


def array(events, slots, objective_function=None):
    """Compute the ILP solution and return it in array form

     Parameters
    ----------
        events : list or tuple
            of resources.Event instances
        slots : list or tuple
            of resources.Slot instances
        objective_function : callable
            from lp_problem.objective_functions

    Returns
    -------
        np.array
            an E by S array (X) where E is the number of events and S the
            number of slots.
            Xij is 1 if event i is scheduled in slot j and zero otherwise

            e.g. 3 events, 7 slots and a solution where:
                event 0 is scheduled in slot 1
                event 1 is scheduled in slot 4
                event 2 is scheduled in slot 5

            the resulting array would be:
                [[0, 1, 0, 0, 0, 0, 0],
                 [0, 0, 0, 0, 1, 0, 0],
                 [0, 0, 0, 0, 0, 1, 0]]
    """
    return solution_to_array(
        solution(events, slots, objective_function),
        events, slots
    )


def schedule(events, slots, objective_function=None, **kwargs):
    """Compute the ILP solution and return it in schedule form

     Parameters
    ----------
        events : list or tuple
            of resources.Event instances
        slots : list or tuple
            of resources.Slot instances
        objective_function : callable
            from lp_problem.objective_functions

    Returns
    -------
        Generator
            of tuples of instances of resources.ScheduledItem
    """
    return solution_to_schedule(
        solution(events, slots, objective_function, **kwargs),
        events, slots
    )


# Functions to convert the schedule from one form to another

def solution_to_array(solution, events, slots):
    array = np.zeros((len(events), len(slots)))
    for item in solution:
        array[item[0], item[1]] = 1
    return array


def solution_to_schedule(solution, events, slots):
    return (
        ScheduledItem(
            event=events[item[0]],
            slot=slots[item[1]]
        )
        for item in solution
    )


def schedule_to_array(schedule, events, slots):
    array = np.zeros((len(events), len(slots)))
    for item in schedule:
        array[events.index(item.event), slots.index(item.slot)] = 1
    return array


def array_to_schedule(array, events, slots):
    scheduled = np.transpose(np.nonzero(array))
    return (
        ScheduledItem(event=events[item[0]], slot=slots[item[1]])
        for item in scheduled
    )



from typing import NamedTuple, Sequence, Dict, Iterable, List
from datetime import datetime


class Slot(NamedTuple):
    venue: str
    starts_at: datetime
    duration: int
    capacity: int
    session: str


class Event(NamedTuple):
    name: str
    duration: int
    tags: List[str]
    unavailability: List
    demand: int


class ScheduledItem(NamedTuple):
    event: Event
    slot: Slot


class Shape(NamedTuple):
    """Represents the shape of a 2 dimensional array of events and slots"""
    events: int
    slots: int


class Constraint(NamedTuple):
    label: str
    condition: bool



from setuptools import setup

setup(
    name='conference-scheduler',
    version='1.0.0',
    packages=[
        'conference_scheduler',
    ],
    url='https://github.com/PyconUK/ConferenceScheduler',
    license='MIT',
    author='Owen Campbell, Vince Knight',
    author_email='owen.campbell@tanti.org.uk',
    description='A Python tool to assist the task of scheduling a conference',
    install_requires=['pulp', 'numpy'],
    setup_requires=['pytest-runner'],
    tests_require=['pytest', 'pytest-pep8'],
    classifiers=[
        'Programming Language :: Python :: 3.6',
    ],
    python_requires='>=3.6',
)
